##
# File: ExDbWorkflow.py
# Date: 17-Dec-2019  jdw
#
#  Workflow wrapper  --  exchange database loading utilities --
#
#  Updates:
#   2-Mar-2023 dwp Add "numProc" parameter to 'upd_ref_seq' operation methods
#   9-Mar-2023 dwp Lower refChunkSize to 10 (UniProt API having trouble streaming XML responses)
#  25-Apr-2024 dwp Add arguments and methods to support CLI usage from weekly-update workflow
#  20-Aug-2024 dwp Add LoadTargetCofactors step; change name of UpdateTargetsCofactors step to UpdateTargetsData
##
__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import logging
import os

from rcsb.db.mongo.DocumentLoader import DocumentLoader
from rcsb.db.utils.TimeUtil import TimeUtil
from rcsb.utils.config.ConfigUtil import ConfigUtil
from rcsb.utils.dictionary.DictMethodResourceProvider import DictMethodResourceProvider
from rcsb.utils.dictionary.DictMethodResourceCacheWorkflow import DictMethodResourceCacheWorkflow
from rcsb.utils.dictionary.NeighborInteractionWorkflow import NeighborInteractionWorkflow
from rcsb.workflow.targets.ProteinTargetSequenceExecutionWorkflow import ProteinTargetSequenceExecutionWorkflow
from rcsb.workflow.chem.ChemCompImageWorkflow import ChemCompImageWorkflow
from rcsb.workflow.chem.ChemCompFileWorkflow import ChemCompFileWorkflow
from rcsb.exdb.chemref.ChemRefEtlWorker import ChemRefEtlWorker
from rcsb.exdb.seq.ReferenceSequenceAnnotationAdapter import ReferenceSequenceAnnotationAdapter
from rcsb.exdb.seq.ReferenceSequenceAnnotationProvider import ReferenceSequenceAnnotationProvider
from rcsb.exdb.seq.UniProtCoreEtlWorker import UniProtCoreEtlWorker
from rcsb.exdb.tree.TreeNodeListWorker import TreeNodeListWorker
from rcsb.exdb.utils.ObjectTransformer import ObjectTransformer
from rcsb.exdb.wf.EntryInfoEtlWorkflow import EntryInfoEtlWorkflow
from rcsb.exdb.wf.GlycanEtlWorkflow import GlycanEtlWorkflow
from rcsb.exdb.wf.PubChemEtlWorkflow import PubChemEtlWorkflow

logger = logging.getLogger(__name__)


class ExDbWorkflow(object):
    def __init__(self, **kwargs):
        #  Configuration Details
        self.__configPath = kwargs.get("configPath", "exdb-config-example.yml")
        self.__configName = kwargs.get("configName", "site_info_remote_configuration")
        self.__mockTopPath = kwargs.get("mockTopPath", None)
        self.__cfgOb = ConfigUtil(configPath=self.__configPath, defaultSectionName=self.__configName, mockTopPath=self.__mockTopPath)
        #
        self.__cachePath = kwargs.get("cachePath", ".")
        self.__cachePath = os.path.abspath(self.__cachePath)
        self.__debugFlag = kwargs.get("debugFlag", False)
        if self.__debugFlag:
            logger.setLevel(logging.DEBUG)
        #
        # Rebuild cache (default False)
        rebuildCache = kwargs.get("rebuildCache", False)
        self.__useCache = not rebuildCache
        providerTypeExcludeL = kwargs.get("providerTypeExcludeL", None)
        restoreUseGit = kwargs.get("restoreUseGit", True)
        restoreUseStash = kwargs.get("restoreUseStash", True)
        self.__cacheStatus = True
        if rebuildCache:
            logger.info("Rebuilding cache %r", rebuildCache)
            self.__cacheStatus = self.buildResourceCache(
                rebuildCache=rebuildCache,
                providerTypeExcludeL=providerTypeExcludeL,
                restoreUseStash=restoreUseStash,
                restoreUseGit=restoreUseGit,
            )
            logger.info("Cache status if %r", self.__cacheStatus)
            if not self.__cacheStatus:
                logger.error("Failed to rebuild CACHE in ExDBWorkflow")

    def load(self, op, **kwargs):
        logger.info("Starting operation %r\n", op)
        #
        # argument processing
        if op not in ["etl_tree_node_lists", "etl_chemref", "etl_uniprot_core", "upd_ref_seq", "upd_ref_seq_comp_models", "refresh_pubchem"]:
            logger.error("Unsupported operation %r - exiting", op)
            return False
        if not self.__cacheStatus:
            raise ValueError("Failed to rebuild CACHE in ExDBWorkflow - exiting")
        try:
            # test mode and UniProt accession primary match minimum count for doReferenceSequenceUpdate()
            testMode = kwargs.get("testMode", False)
            minMatchPrimaryPercent = kwargs.get("minMatchPrimaryPercent", None)
            minMissing = kwargs.get("minMissing", 0)
            #
            readBackCheck = kwargs.get("readBackCheck", True)
            numProc = int(kwargs.get("numProc", 1))
            chunkSize = int(kwargs.get("chunkSize", 10))
            maxStepLength = int(kwargs.get("maxStepLength", 500))
            refChunkSize = int(kwargs.get("refChunkSize", 10))
            documentLimit = kwargs.get("documentLimit", None)
            documentLimit = int(documentLimit) if documentLimit else None
            loadType = kwargs.get("loadType", "full")  # or replace
            dbType = kwargs.get("dbType", "mongo")
            tU = TimeUtil()
            dataSetId = kwargs.get("dataSetId") if "dataSetId" in kwargs else tU.getCurrentWeekSignature()
            #  Rebuild or reuse reference sequence cache
            rebuildSequenceCache = kwargs.get("rebuildSequenceCache", False)
            useSequenceCache = not rebuildSequenceCache
            #
            useFilteredLists = kwargs.get("useFilteredLists", False)

        except Exception as e:
            logger.exception("Argument or configuration processing failing with %s", str(e))
            return False
        #
        okS = ok = False
        if dbType == "mongo":
            if op == "etl_tree_node_lists":
                rhw = TreeNodeListWorker(
                    self.__cfgOb,
                    self.__cachePath,
                    numProc=numProc,
                    chunkSize=chunkSize,
                    maxStepLength=maxStepLength,
                    documentLimit=documentLimit,
                    verbose=self.__debugFlag,
                    readBackCheck=readBackCheck,
                    useCache=self.__useCache,
                    useFilteredLists=useFilteredLists,
                )
                ok = rhw.load(dataSetId, loadType=loadType)
                okS = self.loadStatus(rhw.getLoadStatus(), readBackCheck=readBackCheck)

            elif op == "etl_chemref":
                crw = ChemRefEtlWorker(
                    self.__cfgOb,
                    self.__cachePath,
                    numProc=numProc,
                    chunkSize=chunkSize,
                    maxStepLength=maxStepLength,
                    documentLimit=documentLimit,
                    verbose=self.__debugFlag,
                    readBackCheck=readBackCheck,
                    useCache=self.__useCache,
                )
                ok = crw.load(dataSetId, extResource="DrugBank", loadType=loadType)
                okS = self.loadStatus(crw.getLoadStatus(), readBackCheck=readBackCheck)

            elif op == "etl_uniprot_core":
                crw = UniProtCoreEtlWorker(
                    self.__cfgOb,
                    self.__cachePath,
                    numProc=numProc,
                    chunkSize=chunkSize,
                    maxStepLength=maxStepLength,
                    documentLimit=documentLimit,
                    verbose=self.__debugFlag,
                    readBackCheck=readBackCheck,
                    useCache=self.__useCache,
                )
                ok = crw.load(dataSetId, extResource="UniProt", loadType=loadType)
                okS = self.loadStatus(crw.getLoadStatus(), readBackCheck=readBackCheck)

            elif op == "upd_ref_seq":
                databaseName = "pdbx_core"
                collectionName = "pdbx_core_polymer_entity"
                polymerType = "Protein"
                ok = self.doReferenceSequenceUpdate(
                    databaseName,
                    collectionName,
                    polymerType,
                    fetchLimit=documentLimit,
                    useSequenceCache=useSequenceCache,
                    testMode=testMode,
                    minMatchPrimaryPercent=minMatchPrimaryPercent,
                    minMissing=minMissing,
                    refChunkSize=refChunkSize,
                    numProc=numProc
                )
                okS = ok
            # elif op == "upd_ref_seq_comp_models":
            #     databaseName = "pdbx_comp_model_core"
            #     collectionName = "pdbx_comp_model_core_polymer_entity"
            #     polymerType = "Protein"
            #     ok = self.doReferenceSequenceUpdate(
            #         databaseName,
            #         collectionName,
            #         polymerType,
            #         fetchLimit=documentLimit,
            #         useSequenceCache=useSequenceCache,
            #         testMode=testMode,
            #         minMatchPrimaryPercent=minMatchPrimaryPercent,
            #         minMissing=minMissing,
            #         refChunkSize=refChunkSize,
            #     )
            #     okS = ok
        #
        logger.info("Completed operation %r with status %r\n", op, ok and okS)
        return ok and okS

    def loadStatus(self, statusList, readBackCheck=True):
        ret = False
        try:
            dl = DocumentLoader(self.__cfgOb, self.__cachePath, "MONGO_DB", numProc=1, chunkSize=2, documentLimit=None, verbose=False, readBackCheck=readBackCheck)
            #
            sectionName = "data_exchange_configuration"
            databaseName = self.__cfgOb.get("DATABASE_NAME", sectionName=sectionName)
            collectionName = self.__cfgOb.get("COLLECTION_UPDATE_STATUS", sectionName=sectionName)
            ret = dl.load(databaseName, collectionName, loadType="append", documentList=statusList, indexAttributeList=["update_id", "database_name", "object_name"], keyNames=None)
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ret

    def buildResourceCache(self, rebuildCache=False, providerTypeExcludeL=None, restoreUseStash=True, restoreUseGit=True):
        """Generate and cache resource dependencies."""
        ret = False
        try:
            # First make sure the CACHE directory exists
            if not os.path.isdir(self.__cachePath):
                logger.info("Cache directory %s doesn't exist. Creating it", self.__cachePath)
                os.makedirs(self.__cachePath)
            else:
                logger.info("Cache directory %s already exists.", self.__cachePath)

            # Now build the cache
            useCache = not rebuildCache
            rP = DictMethodResourceProvider(
                self.__cfgOb,
                configName=self.__configName,
                cachePath=self.__cachePath,
                restoreUseStash=restoreUseStash,
                restoreUseGit=restoreUseGit,
                providerTypeExcludeL=providerTypeExcludeL,
            )
            ret = rP.cacheResources(useCache=useCache, doBackup=False)
            logger.info("useCache %r cache reload status (%r)", useCache, ret)

        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return ret

    def doReferenceSequenceUpdate(
        self,
        databaseName,
        collectionName,
        polymerType,
        fetchLimit=None,
        useSequenceCache=False,
        testMode=False,
        minMatchPrimaryPercent=None,
        minMissing=0,
        refChunkSize=10,
        numProc=2,
        **kwargs
    ):
        try:
            _ = kwargs
            _ = testMode
            ok = False
            # -------
            rsaP = ReferenceSequenceAnnotationProvider(
                self.__cfgOb, databaseName, collectionName, polymerType, useCache=useSequenceCache, cachePath=self.__cachePath, maxChunkSize=refChunkSize, numProc=numProc
            )
            ok = rsaP.testCache(minMatchPrimaryPercent=minMatchPrimaryPercent, minMissing=minMissing)
            if ok:
                logger.info("Cached reference data count is %d", rsaP.getRefDataCount())
                rsa = ReferenceSequenceAnnotationAdapter(rsaP)
                obTr = ObjectTransformer(self.__cfgOb, objectAdapter=rsa)
                ok = obTr.doTransform(
                    databaseName=databaseName, collectionName=collectionName, fetchLimit=fetchLimit, selectionQuery={"entity_poly.rcsb_entity_polymer_type": polymerType}
                )
            else:
                logger.error("Reference sequence data cache build failing")
                return False
            return ok
        except Exception as e:
            logger.exception("Failing with %s", str(e))
        return False

    def generateCcdFiles(self, op, **kwargs):
        logger.info("Starting operation %r\n", op)
        #
        # argument processing
        if op not in [
            "ccd_img_gen",
            "ccd_file_gen",
        ]:
            logger.error("Unsupported operation %r - exiting", op)
            return False
        try:
            ccOutputPath = kwargs.get("ccOutputPath", None)
            ccCachePath = kwargs.get("ccCachePath", None)
            licenseFilePath = kwargs.get("licenseFilePath", None)
            ccUrlTarget = kwargs.get("ccUrlTarget", None)
            birdUrlTarget = kwargs.get("birdUrlTarget", None)
            #
        except Exception as e:
            logger.exception("Argument or configuration processing failing with %s", str(e))
            return False
        #
        ok = False
        if op == "ccd_img_gen":
            logger.info("Generating CCD 2D images ...")
            cciWf = ChemCompImageWorkflow(
                imagePath=ccOutputPath,
                cachePath=ccCachePath,
                licenseFilePath=licenseFilePath,
                ccUrlTarget=ccUrlTarget,
                birdUrlTarget=birdUrlTarget,
            )
            ok = cciWf.testCache()
            logger.info("CCD image generation setup status %r", ok)
            #
            ok = cciWf.makeImages() and ok
            logger.info("CCD image generation status %r", ok)
        #
        elif op == "ccd_file_gen":
            logger.info("Generating SDF and Mol2 files from CCD...")
            # CCD ideal coordinates
            ccfWf = ChemCompFileWorkflow(
                fileDirPath=ccOutputPath,
                cachePath=ccCachePath,
                licenseFilePath=licenseFilePath,
                ccUrlTarget=ccUrlTarget,
                birdUrlTarget=birdUrlTarget,
                molBuildType="ideal-xyz"
            )
            ok1 = ccfWf.testCache()
            logger.info("CCD ideal coordinates generation setup status %r", ok1)
            #
            ok1 = ccfWf.makeFiles(fmt="sdf") and ok1
            logger.info("CCD ideal file sdf generation status %r", ok1)
            ok1 = ccfWf.makeFiles(fmt="mol2") and ok1
            logger.info("CCD ideal file mol2 generation status %r", ok1)

            # CCD model coordinates
            ccfWf = ChemCompFileWorkflow(
                fileDirPath=ccOutputPath,
                cachePath=ccCachePath,
                licenseFilePath=licenseFilePath,
                ccUrlTarget=ccUrlTarget,
                birdUrlTarget=birdUrlTarget,
                molBuildType="model-xyz"
            )
            ok2 = ccfWf.testCache()
            logger.info("CCD model coordinates generation setup status %r", ok2)
            #
            ok2 = ccfWf.makeFiles(fmt="sdf") and ok2
            logger.info("CCD model file sdf generation status %r", ok2)
            ok2 = ccfWf.makeFiles(fmt="mol2") and ok2
            logger.info("CCD model file mol2 generation status %r", ok2)
            #
            ok = ok1 and ok2
        #
        logger.info("Completed operation %r with status %r\n", op, ok)
        if not ok:
            logger.error("%r FAILED with status %s", op, ok)
            raise ValueError("%r FAILED. Check the loader log for details." % op)

        return ok

    def buildExdbResource(self, op, **kwargs):
        logger.info("Starting operation %r\n", op)
        #
        # argument processing
        if op not in [
            "upd_neighbor_interactions",
            "upd_uniprot_taxonomy",
            "upd_targets",
            "load_target_cofactors",
            "upd_pubchem",
            "upd_entry_info",
            "upd_glycan_idx",
            "upd_resource_stash",
        ]:
            logger.error("Unsupported operation %r - exiting", op)
            return False
        try:
            numProc = int(kwargs.get("numProc", 1))
            chunkSize = int(kwargs.get("chunkSize", 10))
            documentLimit = kwargs.get("documentLimit", None)
            documentLimit = int(documentLimit) if documentLimit else None
            dbType = kwargs.get("dbType", "mongo")
            rebuildAllNeighborInteractions = kwargs.get("rebuildAllNeighborInteractions", False)
            incrementalUpdate = not rebuildAllNeighborInteractions
            ccFileNamePrefix = kwargs.get("ccFileNamePrefix", "cc-full")
            ccUrlTarget = kwargs.get("ccUrlTarget", None)
            birdUrlTarget = kwargs.get("birdUrlTarget", None)
            rebuildChemIndices = kwargs.get("rebuildChemIndices", True)
            #
        except Exception as e:
            logger.exception("Argument or configuration processing failing with %s", str(e))
            return False
        #
        ok = False
        if dbType == "mongo":
            if op == "upd_neighbor_interactions":
                logger.info("Starting workflow NeighborInteractionWorkflow.buildResourceCache()")
                niWf = NeighborInteractionWorkflow(
                    configPath=self.__configPath,
                    mockTopPath=self.__mockTopPath,
                    configName=self.__configName,
                    cachePath=self.__cachePath,
                    numProc=numProc,
                    chunkSize=chunkSize,
                )
                ok = niWf.update(incremental=incrementalUpdate)
                logger.info("UpdateNeighborInteraction status %r", ok)
                ok = niWf.backup() and ok
                logger.info("UpdateNeighborInteraction backup status %r", ok)
            #
            elif op == "upd_uniprot_taxonomy":
                logger.info("Starting workflow ProteinTargetSequenceExecutionWorkflow (full)")
                ptsWf = ProteinTargetSequenceExecutionWorkflow(
                    configPath=self.__configPath,
                    mockTopPath=self.__mockTopPath,
                    configName=self.__configName,
                    cachePath=self.__cachePath,
                )
                ok = ptsWf.cacheTaxonomy()
                logger.info("cacheTaxonomy status %r", ok)
                ok = ptsWf.updateUniProtTaxonomy() and ok
                logger.info("updateUniProtTaxonomy status %r", ok)
            #
            elif op == "upd_targets":
                logger.info("Starting UpdateTargetsData")
                ptsWf = ProteinTargetSequenceExecutionWorkflow(
                    configPath=self.__configPath,
                    mockTopPath=self.__mockTopPath,
                    configName=self.__configName,
                    cachePath=self.__cachePath,
                )
                ok = ptsWf.fetchProteinEntityData()
                logger.info("fetchProteinEntityData status %r", ok)
                ok = ptsWf.fetchChemicalReferenceMappingData() and ok
                logger.info("fetchChemicalReferenceMappingData status %r", ok)
                ok = ptsWf.fetchLigandNeighborMappingData() and ok
                logger.info("fetchLigandNeighborMappingData status %r", ok)
                ok = ptsWf.exportFasta(reloadPharos=False, fromDbPharos=False) and ok
                logger.info("exportFasta status %r", ok)
                ok = ptsWf.createSearchDatabases() and ok
                logger.info("createSearchDatabases status %r", ok)
                ok = ptsWf.searchDatabases() and ok
                logger.info("searchDatabases status %r", ok)
                ok = ptsWf.buildFeatures() and ok
                logger.info("buildFeatures status %r", ok)
                ok = ptsWf.buildActivityData() and ok
                logger.info("buildActivityData status %r", ok)
                ok = ptsWf.buildCofactorData() and ok
                logger.info("buildCofactorData status %r", ok)
                ptsWf.resourceCheck()
            #
            elif op == "load_target_cofactors":
                logger.info("Starting LoadTargetCofactors")
                ptsWf = ProteinTargetSequenceExecutionWorkflow(
                    configPath=self.__configPath,
                    mockTopPath=self.__mockTopPath,
                    configName=self.__configName,
                    cachePath=self.__cachePath,
                )
                ok = ptsWf.loadTargetCofactorData()
                logger.info("loadTargetCofactorData status %r", ok)
                ptsWf.resourceCheck()
            #
            elif op == "upd_pubchem":
                #  -- Update local chemical indices and  create PubChem mapping index ---
                logger.info("Starting workflow PubChemEtlWorkflow")
                pcewP = PubChemEtlWorkflow(
                    configPath=self.__configPath,
                    mockTopPath=self.__mockTopPath,
                    configName=self.__configName,
                    cachePath=self.__cachePath,
                )
                logger.info("Starting workflow PubChemEtlWorkflow.updateMatchedIndex()")
                ok1 = pcewP.updateMatchedIndex(
                    rebuildChemIndices=rebuildChemIndices,
                    ccUrlTarget=ccUrlTarget,
                    birdUrlTarget=birdUrlTarget,
                    ccFileNamePrefix=ccFileNamePrefix,
                    # numProc=numProc,  # Leave at default of 2, due to throttling at PubChem
                    # chunkSize=chunkSize,  # need to propagate this through
                )
                logger.info("updateMatchedIndex status %r", ok1)
                #
                logger.info("Starting workflow PubChemEtlWorkflow.updateMatchedData()")
                ok2 = pcewP.updateMatchedData(numProc=numProc)
                logger.info("updateMatchedData status %r", ok2)
                ok = ok1 and ok2
            #
            elif op == "upd_entry_info":
                logger.info("Starting workflow EntryInfoEtlWorkflow.update()")
                ewf = EntryInfoEtlWorkflow(
                    configPath=self.__configPath,
                    mockTopPath=self.__mockTopPath,
                    configName=self.__configName,
                    cachePath=self.__cachePath,
                )
                ok = ewf.update()
            #
            elif op == "upd_glycan_idx":
                logger.info("Starting workflow GlycanEtlWorkflow.updateMatchedIndex()")
                gwf = GlycanEtlWorkflow(
                    configPath=self.__configPath,
                    mockTopPath=self.__mockTopPath,
                    configName=self.__configName,
                    cachePath=self.__cachePath,
                )
                ok = gwf.updateMatchedIndex()
            #
            elif op == "upd_resource_stash":
                logger.info("Starting workflow DictMethodResourceCacheWorkflow.buildResourceCache()")
                dmWf = DictMethodResourceCacheWorkflow(
                    configPath=self.__configPath,
                    mockTopPath=self.__mockTopPath,
                    configName=self.__configName,
                    cachePath=self.__cachePath,
                )
                ok = dmWf.buildResourceCache()
        #
        logger.info("Completed operation %r with status %r\n", op, ok)
        if not ok:
            logger.error("%r FAILED with status %s", op, ok)
            raise ValueError("%r FAILED. Check the loader log for details." % op)

        return ok



##
# File: ExDbExec.py
# Date: 22-Apr-2019  jdw
#
#  Execution wrapper  --  for extract and load operations -
#
#  Updates:
#   4-Sep-2019 jdw add Tree and Drugbank loaders
#  14-Feb-2020 jdw change over to ReferenceSequenceAnnotationProvider/Adapter
#   9-Mar-2023 dwp Lower refChunkSize to 10 (UniProt API having trouble streaming XML responses)
#  25-Apr-2024 dwp Add arguments and logic to support CLI usage from weekly-update workflow;
#                  Add support for logging output to a specific file
#  20-Aug-2024 dwp Add load_target_cofactors operation; change name of upd_targets_cofactors to upd_targets
#  22-Oct-2024 dwp Add ccd_img_gen and ccd_file_gen operations
#                  (latter will only be used briefly, as will stop generating SDF and Mol2 files in Dec 2024)
##
__docformat__ = "google en"
__author__ = "John Westbrook"
__email__ = "jwest@rcsb.rutgers.edu"
__license__ = "Apache 2.0"

import os
import sys
import argparse
import logging

from rcsb.utils.config.ConfigUtil import ConfigUtil
from rcsb.exdb.wf.ExDbWorkflow import ExDbWorkflow

HERE = os.path.abspath(os.path.dirname(__file__))
TOPDIR = os.path.dirname(os.path.dirname(os.path.dirname(HERE)))

# logging.basicConfig(level=logging.INFO, format="%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s", stream=sys.stdout)
logger = logging.getLogger()


def main():
    parser = argparse.ArgumentParser()
    #
    parser.add_argument(
        "--op",
        default=None,
        required=True,
        help="Loading operation to perform",
        choices=[
            "etl_chemref",  # ETL integrated chemical reference data
            "etl_uniprot_core",  # ETL UniProt core reference data
            "etl_tree_node_lists",  # ETL tree node lists
            "upd_ref_seq",  # Update reference sequence assignments
            "upd_neighbor_interactions",
            "upd_uniprot_taxonomy",
            "upd_targets",
            "load_target_cofactors",
            "upd_pubchem",
            "upd_entry_info",
            "upd_glycan_idx",
            "upd_resource_stash",
            "ccd_img_gen",
            "ccd_file_gen",
        ]
    )
    parser.add_argument(
        "--load_type",
        default="full",
        help="Type of load ('full' for complete and fresh single-worker load, 'replace' for incremental and multi-worker load)",
        choices=["full", "replace"],
    )
    #
    parser.add_argument("--config_path", default=None, help="Path to configuration options file")
    parser.add_argument("--config_name", default="site_info_remote_configuration", help="Configuration section name")
    parser.add_argument("--cache_path", default=None, help="Cache path for resource files")
    parser.add_argument("--num_proc", default=2, help="Number of processes to execute (default=2)")
    parser.add_argument("--chunk_size", default=10, help="Number of files loaded per process")
    parser.add_argument("--max_step_length", default=500, help="Maximum subList size (default=500)")
    parser.add_argument("--db_type", default="mongo", help="Database server type (default=mongo)")
    parser.add_argument("--document_limit", default=None, help="Load document limit for testing")
    #
    parser.add_argument("--rebuild_cache", default=False, action="store_true", help="Rebuild cached resource files")
    parser.add_argument("--rebuild_sequence_cache", default=False, action="store_true", help="Rebuild cached resource files for reference sequence updates")
    parser.add_argument("--provider_types_exclude", default=None, help="Resource provider types to exclude")
    parser.add_argument("--use_filtered_tax_list", default=False, action="store_true", help="Use filtered list for taxonomy tree loading")
    parser.add_argument("--disable_read_back_check", default=False, action="store_true", help="Disable read back check on all documents")
    parser.add_argument("--debug", default=False, action="store_true", help="Turn on verbose logging")
    parser.add_argument("--mock", default=False, action="store_true", help="Use MOCK repository configuration for testing")
    parser.add_argument("--log_file_path", default=None, help="Path to runtime log file output.")
    #
    # Arguments specific for op == 'upd_ref_seq'
    parser.add_argument("--ref_chunk_size", default=10, help="Max chunk size for reference sequence updates (for op 'upd_ref_seq')")
    parser.add_argument("--min_missing", default=0, help="Minimum number of allowed missing reference sequences (for op 'upd_ref_seq')")
    parser.add_argument("--min_match_primary_percent", default=None, help="Minimum reference sequence match percentage (for op 'upd_ref_seq')")
    parser.add_argument("--test_mode", default=False, action="store_true", help="Test mode for reference sequence updates (for op 'upd_ref_seq')")
    #
    # Arguments specific for op == 'ccd_img_gen' or 'ccd_file_gen'
    parser.add_argument("--cc_output_path", default=None, help="The base local directory path where chemical component files (image, coordinates) are written (for op 'ccd_img_gen')")
    parser.add_argument("--cc_cache_path", default=None, help="The base local directory path where chemical component cache data are written (for op 'ccd_img_gen')")
    parser.add_argument("--oe_license_path", default=None, help="Path to OpenEye license file")
    #
    # Arguments buildExdbResources
    parser.add_argument("--rebuild_all_neighbor_interactions", default=False, action="store_true", help="Rebuild all neighbor interactions from scratch (default is incrementally)")
    parser.add_argument("--cc_file_prefix", default="cc-full", help="File name discriminator for index sets")
    parser.add_argument("--cc_url_target", default=None, help="target url for chemical component dictionary resource file (default: None=all public)")
    parser.add_argument("--bird_url_target", default=None, help="target url for bird dictionary resource file (cc format) (default: None=all public)")
    #
    args = parser.parse_args()
    #
    try:
        op, commonD, loadD = processArguments(args)
    except Exception as err:
        logger.exception("Argument processing problem %s", str(err))
        raise ValueError("Argument processing problem") from err
    #
    #
    # Log input arguments
    loadLogD = {k: v for d in [commonD, loadD] for k, v in d.items() if k != "inputIdCodeList"}
    logger.info("running load op %r on loadLogD %r:", op, loadLogD)
    #
    # Run the operation
    okR = False
    exWf = ExDbWorkflow(**commonD)
    if op in ["etl_chemref", "etl_uniprot_core", "etl_tree_node_lists", "upd_ref_seq"]:
        okR = exWf.load(op, **loadD)
    elif op in [
        "upd_neighbor_interactions",
        "upd_uniprot_taxonomy",
        "upd_targets",
        "load_target_cofactors",
        "upd_pubchem",
        "upd_entry_info",
        "upd_glycan_idx",
        "upd_resource_stash",
    ]:
        okR = exWf.buildExdbResource(op, **loadD)
    elif op in [
        "ccd_img_gen",
        "ccd_file_gen",
    ]:
        okR = exWf.generateCcdFiles(op, **loadD)
    else:
        logger.error("Unsupported op %r", op)
    #
    logger.info("Operation %r completed with status %r", op, okR)
    #
    if not okR:
        logger.error("Operation %r failed with status %r", op, okR)
        raise ValueError("Operation %r failed" % op)


def processArguments(args):
    # Logging details
    logFilePath = args.log_file_path
    debugFlag = args.debug
    if debugFlag:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)
    if logFilePath:
        logDir = os.path.dirname(logFilePath)
        if not os.path.isdir(logDir):
            os.makedirs(logDir)
        handler = logging.FileHandler(logFilePath, mode="a")
        if debugFlag:
            handler.setLevel(logging.DEBUG)
        else:
            handler.setLevel(logging.INFO)
        formatter = logging.Formatter("%(asctime)s [%(levelname)s]-%(module)s.%(funcName)s: %(message)s")
        handler.setFormatter(formatter)
        logger.addHandler(handler)
    #
    # Configuration details
    configPath = args.config_path
    configName = args.config_name
    if not (configPath and configName):
        logger.error("Config path and/or name not provided: %r, %r", configPath, configName)
        raise ValueError("Config path and/or name not provided: %r, %r" % (configPath, configName))
    mockTopPath = os.path.join(TOPDIR, "rcsb", "mock-data") if args.mock else None
    logger.info("Using configuration file %r (section %r)", configPath, configName)
    cfgOb = ConfigUtil(configPath=configPath, defaultSectionName=configName, mockTopPath=mockTopPath)
    cfgObTmp = cfgOb.exportConfig()
    logger.info("Length of config object (%r)", len(cfgObTmp))
    if len(cfgObTmp) == 0:
        logger.error("Missing or access issue for config file %r", configPath)
        raise ValueError("Missing or access issue for config file %r" % configPath)
    else:
        del cfgObTmp
    #
    # Do any additional argument checking
    op = args.op
    if not op:
        raise ValueError("Must supply a value to '--op' argument")
    #
    cachePath = args.cache_path if args.cache_path else "."
    cachePath = os.path.abspath(cachePath)

    if args.db_type != "mongo":
        logger.error("Unsupported database type %r (must be 'mongo')", args.db_type)
        raise ValueError("Unsupported database type %r (must be 'mongo')" % args.db_type)

    # Now collect arguments into dictionaries
    commonD = {
        "configPath": configPath,
        "configName": configName,
        "cachePath": cachePath,
        "mockTopPath": mockTopPath,
        "debugFlag": debugFlag,
        "rebuildCache": args.rebuild_cache,
        "providerTypeExcludeL": args.provider_types_exclude,
    }
    loadD = {
        "loadType": args.load_type,
        "numProc": int(args.num_proc),
        "chunkSize": int(args.chunk_size),
        "maxStepLength": int(args.max_step_length),
        "dbType": args.db_type,
        "documentLimit": int(args.document_limit) if args.document_limit else None,
        "readBackCheck": not args.disable_read_back_check,
        "rebuildSequenceCache": args.rebuild_sequence_cache,
        "useFilteredLists": args.use_filtered_tax_list,
        "refChunkSize": int(args.ref_chunk_size),
        "minMissing": int(args.min_missing),
        "minMatchPrimaryPercent": float(args.min_match_primary_percent) if args.min_match_primary_percent else None,
        "testMode": args.test_mode,
        "rebuildAllNeighborInteractions": args.rebuild_all_neighbor_interactions,
        "ccFileNamePrefix": args.cc_file_prefix,
        "ccUrlTarget": args.cc_url_target,
        "birdUrlTarget": args.bird_url_target,
        "ccOutputPath": args.cc_output_path,
        "ccCachePath": args.cc_cache_path,
        "licenseFilePath": args.oe_license_path,
    }

    return op, commonD, loadD


if __name__ == "__main__":
    try:
        main()
    except Exception as e:
        logger.exception("Run failed %s", str(e))
        sys.exit(1)
